/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#ifndef DATA_STRUCTS_H
#define DATA_STRUCTS_H

#pragma warning(disable : 4512)
// this is the warning for "<class>: assignment operator could not be generated"

#include "err_and_warn.h"
#include "Statistic.h"
namespace rogus
{


//--------------------------------------------------------------------------------------------------

class len_monitor : Statistic // 1
{
	const int size;  // maximum possible length of queue
	int len, max;    // current length of queue, maximum length attained
	const char *const name;
public:
   len_monitor(int s, const char *n) : size(s), name(n) { len = max = 0; } // 2
	 
	void clear() { max = 0; }
	void stat(FILE* f) { if (*name) fprintf(f, "%s: %d of %d used\n\n", name, max, size ); }
	
	void inc() { if (++len > max) max = len; }
	void dec() { len--; }
};

/* 1. This allows us to see how much of the queue we are actually using to get an idea of what is 
a safe size for various real-world applications.
2. Null name indicates "do not report usage upon destruction." */

//--------------------------------------------------------------------------------------------------

template<class T>
class c_queue // circular queue
{
	T *const q_bot, *const q_top, *q_beg, *q_end;
	len_monitor mon; // 1
	
	T* next( T* p ) { return p == q_top ? q_bot : p+1; }
		
public:
	c_queue( int s, const char *name );
	
	inline err_ind insertp( T *xp ); // returns err if queue already full, no_err otherwise

	err_ind insert( T x ) { return insertp( &x ); }


	T* headp() { return q_end == q_beg ? NULL : q_beg; }

	void remove_head() { q_beg = next( q_beg ); mon.dec(); }
};

/* 1. Length monitoring may not be 100% accurate if len++ and len-- are not atomic operations, but
who cares since monitoring is only used to get an idea of queue usage. */

//--------------------------------------------------------------------------------------------------

template<class T>
class pool
{
	c_queue<T*> free_list;
	const int size;

protected:
	T *const M;
	
public:
   enum rrv { okay, foreign, excess }; // release() return values

	pool(int s, const char* name);
	T* allocate();	// returns NULL if out of memory
	rrv release(T* m);
};

//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------

template<class T>
c_queue<T>::
c_queue( int s, const char *name ) : mon(s, name),
   q_bot(new T[s+1]), q_top(q_bot + s) // 2
{
	q_beg = q_end = q_bot;
}

/* 2. The queue actually contains 1 more element than needed.  This element functions as a kind of
spacer which allows "full queue" state to be differentiated from "empty queue" state. */

template<class T>
inline
err_ind
c_queue<T>::
insertp( T *xp ) // returns err if queue already full, no_err otherwise
{
	T* n = next(q_end); if ( n == q_beg ) return err;
	*q_end = *xp;
	q_end = n;
	mon.inc();
	return no_err;
}


//--------------------------------------------------------------------------------------------------

template<class T>
pool<T>::
pool(int s, const char* name) : size(s), free_list(s, name), M(new T[s])
{
	for (int i=0; i<s; i++) free_list.insert(M+i);
}

template<class T>
inline
T*
pool<T>::
allocate() // returns NULL if out of memory
{
	T **h = free_list.headp();
	if (h)
	{
		free_list.remove_head();
		return *h;
	}
	else return NULL;
}

template<class T>
inline
pool<T>::rrv
pool<T>::
release(T* m)
{
	if ( m < M || (m - M) > (size-1) ) return foreign;
	err_ind e = free_list.insert(m); if (e) return excess;
   return okay;
}
}
#endif
// Author: Ben Denckla
