/*
** Copyright 1995-1997 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "sched.h"
#include "err_and_warn.h"
#include <process.h>

namespace rogus
{

task*
scheduler::task_pool::
store( const task_data& td )
{
   task* t = allocate(); if (!t) return NULL;
	t->td = td;
	mon.inc();
	return t;
}


/*
A (hopefully correct) description of how (I think) the scheduler works:

Tasks are allocated from the task_pool as they are needed.
Every currently allocated (and therefore currently scheduled) task is linked
together in a linked list. linkedListOfTasks points to the head of this list.
(Of course, if there are no currently scheduled tasks, then linkListOfTasks is
NULL.)

The scheduler thread calls findNextTask() to determine which task should be executed
next. (In other words, find the minimum nwu among all the tasks in the list.)
findNextTask() sets doMeNext to point to the task which should be executed next.
findNextTask() sets doMeNext_prev to the pointer in the linked list which points
to the task identified be doMeNext. (This is so the linked list can be fixed up
when the task is removed.)

Next, the scheduler thread calls waitAndHandle(), which releases the WrappedCS and
then waits for either the amount of time until the next task to elapse, or for
a MIDI input event to occur.

Once the appropriate amount of time has elapsed, waitAndHandle() calls removeAndExecute(),
which, as its name might imply, removes the task indicated by doMeNext from the linked
list, and then executes the task.

A few notes about weird situations that can occur because of multithreadedness:

If a new task is inserted while waitAndHandle() is waiting, then we trick it into
thinking a MIDI input event has occured (this causes no harm, since there are no actual
events in the input queue), which then causes findNextTask() to be re-executed
immediately.

If a task other than the one pointed to by doMeNext is removed while waitAndHandle()
is waiting, then no problem, unless doMeNext_prev is pointing at the chain member of the
task being removed; in this case, doMeNext_prev must be fixed up. If, on the other hand,
the task being removed is the next one to be executed, then we set a special flag,
zappedYourTask, to indicate that the task being waited for no longer exists.
*/

scheduler::
scheduler(user_prefs *in_upp, WrappedCS* in_csp, err_and_warn *in_ewp) :
   tp(1000)
{
   upp = in_upp;
   linkedListOfTasks = NULL;
	csp = in_csp;
   ewp = in_ewp;

   eventHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
  	running = true;
	threadHandle = (HANDLE) _beginthreadex(NULL, 0, threadMain, this, 0, &threadID);
	SetThreadPriority(threadHandle, THREAD_PRIORITY_TIME_CRITICAL);
}

scheduler::
~scheduler()
{
	running = false;
	SetEvent(eventHandle); // wake up scheduler thread; let it see we're not running
	DWORD result = WaitForSingleObject(threadHandle, 1000); // wait for thread to die
	if (result == WAIT_TIMEOUT)
		ewp->w("scheduler thread failed to exit");
	if (!CloseHandle(eventHandle))
		ewp->w("failed to close eventHandle in scheduler cleanup");
	if (!CloseHandle(threadHandle))
		ewp->w("failed to close threadHandle in scheduler cleanup");
}

unsigned WINAPI
scheduler::
threadMain(LPVOID lpvThreadParm)
{
   scheduler* sp = (scheduler*) lpvThreadParm;
	sp->csp->enter();
	while (sp->running)
	{
		sp->findNextTask();
		sp->waitAndHandle(); // releases synchronization while waiting
	}
	sp->csp->leave();
	return 0;
}

void
scheduler::
findNextTask()
{
	if (linkedListOfTasks != NULL) // find out what should be scheduled next
	{
		task* min = (task*) linkedListOfTasks;
		task* i;
		task* min_prev = NULL;
		task* i_prev = NULL;
#ifndef NDEBUG
		if (linkedListOfTasks == NULL)
			ewp->w("sanity check failed in scheduler -- linkedListOfTasks is NULL");
#endif
		for (i = (task*) linkedListOfTasks; i != NULL; i = i->chain)
		{
#ifndef NDEBUG
			if (i == NULL)
				ewp->w("sanity check failed in scheduler -- i is NULL");
#endif
			if (i->td.nwu <= min->td.nwu) // 1
			{
				min = i;
				min_prev = i_prev;
			}
			i_prev = i;
		}
		// 2
		
#ifndef NDEBUG
		if (min == NULL)
			ewp->w("sanity check failed in scheduler -- min is NULL");
#endif
		doMeNext = min;
		doMeNext_prev = (min_prev == NULL) ? &linkedListOfTasks : &(min_prev->chain);
	}
	zappedYourTask = false;
}

/*
1. Explanation behind why this is "<=" instead of "<":
More recently scheduled tasks will appear first, since we add to the head of the queue.
But among tasks which are scheduled for the exact same time, we want to execute first
the one which was scheduled first.

2. At this point, min points to the task which should be scheduled next
min_prev points to the task right before min, or NULL if min == linkedListOfTasks
linkedListOfTasks points to the first task in the queue
i_prev points to the last task in the queue
*/


void
scheduler::
waitAndHandle()
{
	DWORD dwTimeout;
	if (linkedListOfTasks != NULL)
	{
		long delay = doMeNext->td.nwu - oap->time(); // calculate time for wakeup of next task
		dwTimeout = ((delay > 0) ? (delay/2) : 0);
	}
	else
		dwTimeout = INFINITE; // no task scheduled; delay forever (until we get MIDI input)
	csp->leave();
	DWORD result = WaitForSingleObject(eventHandle, dwTimeout);
	csp->enter();
	if (result == WAIT_TIMEOUT)
		removeAndExecute(); // execute scheduled task
	else
		oap->processQueue(); // process MIDI input
}

void
scheduler::
removeAndExecute()
{
	if (zappedYourTask) // this might happen if main thread 
		return;         // removed a scheduled event while we were sleeping

#ifndef NDEBUG
	if (linkedListOfTasks == NULL)
	{
		ewp->w("sanity check failed in scheduler in removeAndExecute() -- linkedListOfTasks is NULL");
		return;
	}
#endif

	task* t = doMeNext;
	task_data& td = doMeNext->td;
	*doMeNext_prev = doMeNext->chain;

	if (upp->statistics_on)
		tp.record_lateness( oap->time() - td.nwu );
	
	// execute task
	bool retval = false;
	switch (td.ty)
	{
		case task_data::fa_t:
		{
			long r_time = td.fa.func( td.fa.arg );
			if (r_time != NEVER)
			{
				td.nwu += r_time;
				retval = true;
			}
			break;
		}
		case task_data::m_t: oap->write( &(td.m) ); break;
		case task_data::on_t:
			oap->write( &(td.m) );
			td.ty = task_data::m_t;
			td.m.d2 = 0;
			td.nwu += td.m.time;
			retval = true;
			break;
	}

	if (!retval)
	{
		tp.mon.dec();
		cr( tp.release( t ) );
	}
	else // implement ZOR
	{
		do_insert(t);
	}
}

err_ind
scheduler::
play_note( const MidiMsg& on, long dur )
{
	oap->write( &on );
	MidiMsg off = on;
	off.d2 = 0;
	task_data td( off, oap->time() + dur );
	return insert(td);
}

err_ind
scheduler::
do_insert(task* t)
{
	t->chain = linkedListOfTasks;
	linkedListOfTasks = t;
	if (GetCurrentThreadId() != threadID) // this hack is so that if we schedule from another thread,
		SetEvent(eventHandle);             // we will recompute which event is scheduled next
	return no_err;
}

err_ind
scheduler::
cancel(task* t)
{
	csp->enter();
	if (t == linkedListOfTasks)
	{
		if (t == doMeNext)
			// indicate that we removed the currently
			// scheduled task
			zappedYourTask = true;
		if (&(t->chain) == doMeNext_prev)
			// We are removing the task which is right before
			// the currently scheduled task, so fix
			// doMeNext_prev to point to the pointer which is
			// pointing to t, which in this case is
			// linkedListOfTasks, since we are at the head of
			// the list.
			doMeNext_prev = &linkedListOfTasks;
		// Since t == linkedListOfTasks, this means that we are
		// removing the head of this list, so need to fix up
		// linkedListOfTasks to point to the new head
		linkedListOfTasks = t->chain;
		tp.mon.dec();
		cr( tp.release(t) );
		csp->leave();
		return no_err;
	}
	task* i;
	task* i_prev = NULL;
	// won't hit first time through loop (when i_prev == NULL) because we
	// know t != linkedListOfTasks (just checked for that above)
	for (i = (task*) linkedListOfTasks; i != NULL; i = i->chain)
	{
		if (i == t)
		{
			if (i == doMeNext)
				// indicate that we removed the currently
				// scheduled task
				zappedYourTask = true;
			if (&(i->chain) == doMeNext_prev)
				// We are removing the task which is right
				// before the currently scheduled task, so fix
				// doMeNext_prev to point to the pointer which
				// is pointing to t.
				doMeNext_prev = &(i_prev->chain);
			i_prev->chain = i->chain; // remove link
			tp.mon.dec();
			cr( tp.release( t ) );
			csp->leave();
			return no_err;
		}
		i_prev = i;
	}
	ewp->w("failed attempt to cancel task -- Event not found!");
	csp->leave();
	return err;
}


void scheduler::print_task_list()
{
	if (linkedListOfTasks != NULL) // find out what should be scheduled next
	{
		task* i;
		for (i = (task*) linkedListOfTasks; i != NULL; i = i->chain)
		{
		}
	}
}

void scheduler::cr( task_pool::rrv retval )
{
   switch (retval)
   {
   case task_pool::okay:
      return;
   case task_pool::foreign:
      ewp->w("You released something not allocated to you");
      return;
   case task_pool::excess:
      ewp->w("You released more times than you allocated");
      return;
   default:
      ewp->w("release() returned something unexpected");
   }
}

}
// Authors: Ben Denckla, Patrick Pelletier
