/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "text_win.h"
#include "Rogus_McBogus.h"
#include "err_and_warn.h"
#include <string.h>
#include <stdio.h> // for vsprintf
#include <process.h> // for _beginthreadex
namespace rogus
{


void
text_win::
my_outfunc(const char* msg, msgt msg_type)
{
	HANDLE hStdout;
	CONSOLE_SCREEN_BUFFER_INFO csbinfo;
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(hStdout, &csbinfo); // save old color
   WORD attrib = FOREGROUND_INTENSITY;
	switch (msg_type)
	{
	case info: attrib |= FOREGROUND_GREEN; break;
	case warn: attrib |= FOREGROUND_GREEN | FOREGROUND_RED; break;
	case err:  attrib |= FOREGROUND_RED; break;
	}
	SetConsoleTextAttribute(hStdout, attrib);
	DWORD numchars = strlen(msg);
	WriteConsole(hStdout, (const void*) msg, numchars, &numchars, NULL);
	SetConsoleTextAttribute(hStdout, csbinfo.wAttributes); // restore old color
}

text_win::
~text_win()
{
	if (inited)
	{
		running = false;
		SetEvent(eventHandle); // wake up output thread; let it see we're not running
		DWORD result = WaitForSingleObject(threadHandle, 1000); // wait for thread to die
		if (result == WAIT_TIMEOUT)
			outfunc("Warning: text_win thread failed to exit\n",warn);
		if (!CloseHandle(eventHandle))
			outfunc("Warning: failed to close eventHandle in text_win cleanup\n",warn);
		if (!CloseHandle(threadHandle))
			outfunc("Warning: failed to close threadHandle in text_win cleanup\n",warn);
		inited = false;
	}
}

unsigned WINAPI
text_win::
threadMain(LPVOID lpvThreadParm)
{
   text_win *twp = (text_win*) lpvThreadParm;
	while (twp->running)
	{
		WaitForSingleObject(twp->eventHandle, INFINITE);
		while (twp->len > 0)
		{
			unit* uu = &twp->units[twp->pos];
			if (twp->len > 1)
			{
				twp->outfunc(uu->buf, uu->msg_type);
				twp->cs.enter();
				twp->pos = (twp->pos + 1) % numunits;
				twp->len--;
				twp->cs.leave();
			}
			else
			{
				twp->cs.enter();
				unit u = *uu;
				twp->pos = (twp->pos + 1) % numunits;
				twp->len--;
				twp->cs.leave();
				twp->outfunc(u.buf, u.msg_type);
			}
		}
	}
	return 0;
}

void
text_win::
init(outfn* func)								
{
	if (!inited)
	{
		pos = 0;
		len = 0;
		UINT threadID;
		eventHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
		running = true;
		threadHandle = (HANDLE) _beginthreadex(NULL, 0, threadMain, this, 0, &threadID);
		SetThreadPriority(threadHandle, THREAD_PRIORITY_LOWEST);
	}
	if (func != NULL)
	{
		outfunc = func;
		inited = true;
		return;
	}
	outfunc = my_outfunc;
	AllocConsole(); // create a console window if we don't already have one
	inited = true;
	// check if we are running under Windows 95
	// if so, don't add scrollbar, because Windows 95 console
	// windows have a serious bug when using scrollbars
	OSVERSIONINFO ovi;
	ovi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if (!GetVersionEx(&ovi))
	{
		pr("GetVersionEx failed");
		return;
	}
	if (ovi.dwPlatformId != VER_PLATFORM_WIN32_NT)
	{
		pr("Note: scrollbar not available under Windows 95\n");
		return;
	}
	// make sure console window has at least 200 lines to scroll back
	HANDLE hStdout;
	CONSOLE_SCREEN_BUFFER_INFO info;
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(hStdout, &info);
	bool dochange = false;
	if (info.dwSize.X < 80)
	{
		info.dwSize.X = 80;
		dochange = true;
	}
	if (info.dwSize.Y < 200)
	{
		info.dwSize.Y = 200;
		dochange = true;
	}
	if (dochange)
		SetConsoleScreenBufferSize(hStdout, info.dwSize);
}

void
text_win::
pr( const char *format, ... )
{
	va_list args;
	va_start( args, format );
	
	vpr( format, args, info );
	
	va_end( args );
}	

void
text_win::
vpr( const char *format, va_list args, msgt msg_type)
{
	if (!inited)
		init(NULL);
	int l = vsprintf( buf, format, args );
	char* where = buf;
	bool prichanged = false;
	int backoff = 1;
	while (l > 0)
	{
		cs.enter();
		unit* u1 = &units[(pos+len)%numunits];
		unit* uu = u1;
		if (len > 0) // try to stuff into previous one if possible
		{
			unit* u2 = &units[(pos+len-1)%numunits];
			if (u2->msg_type == msg_type && (u2->len + l) < bufsize)
				uu = u2;
		}
		if (uu == u1)
		{
			if (len == numunits) // running behind?
			{
				cs.leave();
				SetThreadPriority(threadHandle, THREAD_PRIORITY_TIME_CRITICAL); // bump up priority
				prichanged = true;
				SetEvent(eventHandle); // for good measure
				Sleep(backoff);
				if (backoff < 1024)
					backoff *= 2;
				continue;
			}
			uu->len = 0;
			len++;
		}
		int start = uu->len;
		for (int i = 0; where[i] != 0 && i < bufsize - 1; i++)
		{
			if (where[i] == '\r')
				uu->buf[i+start] = '\n';
			else
				uu->buf[i+start] = where[i];
		}
		uu->buf[i+start] = 0;
		uu->len += i;
		uu->msg_type = msg_type;
		where += i;
		l -= i;
		cs.leave();
	}

	if (prichanged)
		SetThreadPriority(threadHandle, THREAD_PRIORITY_LOWEST);

	SetEvent(eventHandle);
}
}
// Authors: Ben Denckla, Pete Rice, Patrick Pelletier
