/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "ScoreCreationInfo.h"
#include "Score_Players.h"

namespace rogus
{


ScoreCreationInfo::ScoreCreationInfo (Rogus_McBogus* in_r, FiledataPtr f) :
   track_included(f->getNumTracks()), track_node(f->getNumTracks()), track_name(f->getNumTracks())
{
   r = in_r;
	fd = f;
	division = f->getDivision();
	int i;
	for (i = 0; i < f->getNumTracks(); i++)
		track_included[i] = false;

	calculate_track_names();

	num_msgs = 0;
	num_msgs_allocated = 30;
	msgs = (MidiMsg**) malloc(sizeof(MidiMsg*) * num_msgs_allocated);
	if (msgs == NULL)
      { r->ew.e("malloc failed"); throw rx(); }

	num_funcs = 0;
	num_funcs_allocated = 30;
	funcs = (FuncEl**) malloc(sizeof(FuncEl*) * num_funcs_allocated);
	if (funcs == NULL)
		{ r->ew.e("malloc failed"); throw rx(); }
}

ScoreCreationInfo::ScoreCreationInfo (Rogus_McBogus* in_r, uint2 div) :
   track_included(0), track_node(0), track_name(0)
{
   r = in_r;
	fd = NULL;
	division = div;

	num_msgs = 0;
	num_msgs_allocated = 30;
	msgs = (MidiMsg**) malloc(sizeof(MidiMsg*) * num_msgs_allocated);
	if (msgs == NULL)
		{ r->ew.e("malloc failed"); throw rx(); }

	num_funcs = 0;
	num_funcs_allocated = 30;
	funcs = (FuncEl**) malloc(sizeof(FuncEl*) * num_funcs_allocated);
	if (funcs == NULL)
		{ r->ew.e("malloc failed"); throw rx(); }
}

ScoreCreationInfo::~ScoreCreationInfo ()
{
	int i;
	for (i = 0; i < num_msgs; i++)
		delete msgs[i];
	free(msgs);
	for (i = 0; i < num_funcs; i++)
		delete funcs[i];
	free(funcs);
}

err_ind ScoreCreationInfo::addTrack(const char* t, ioref n)
{
	UINT i;
	for (i = 0; i < track_name.length; i++)
		if (*(track_name[i]) == t)
			break;
	if (i == track_name.length) // track doesn't exist
	{
		r->ew.w("track does not exist: %s", t);
		return err;
	}
	if (track_included[i])
		return no_err; // track already added
	track_included[i] = true;
	track_node[i] = n;
	return no_err;
}

int ScoreCreationInfo::addRemainingTracks(ioref n)
{
	int result = 0;
	UINT i;
	for (i = 0; i < track_name.length; i++)
		if (!track_included[i])
		{
			track_included[i] = true;
			track_node[i] = n;
			result++;
		}
	return result;
}

int ScoreCreationInfo::addRemainingTracksAndLookupNodenames()
{
	int result = 0;
	UINT i;
	for (i = 0; i < track_name.length; i++)
		if (!track_included[i])
		{
			ioref n = r->o.open_output(*(track_name[i]));
			track_included[i] = true;
			track_node[i] = n;
			result++;
		}
	return result;
}

void ScoreCreationInfo::addMidiMsg(const MidiMsg& m)
{
	if (num_msgs == num_msgs_allocated)
	{
		num_msgs_allocated *= 2;
		msgs = (MidiMsg**) realloc(msgs, sizeof(MidiMsg*) * num_msgs_allocated);
		if (msgs == NULL)
			{ r->ew.e("realloc failed"); throw rx(); }
	}
	msgs[num_msgs++] = new MidiMsg(m);
}

void ScoreCreationInfo::addFunction(void (*func)(void*, ScorePlayer*), void* v, ScheduleTime t)
{
	if (num_funcs == num_funcs_allocated)
	{
		num_funcs_allocated *= 2;
		funcs = (FuncEl**) realloc(funcs, sizeof(FuncEl*) * num_funcs_allocated);
		if (funcs == NULL)
         { r->ew.e("realloc failed"); throw rx(); }
	}
	FuncEl* tmp = new FuncEl;
	tmp->func = func;
	tmp->v = v;
	tmp->time = t;
	funcs[num_funcs++] = tmp;
}

void ScoreCreationInfo::calculate_track_names()
{
	UINT i;
	for (i = 0; i < track_name.length; i++)
	{
		MidiMsgArrayPtr p1 = fd->getTrack(i);
		UINT j;
		for (j = 0; j < p1->length; j++) // find track name
		{
			MidiMsg* m = &((*p1)[j]);
			if (m->is_meta(meta::txt_seq_or_trk))
			{
				track_name[i] = (m->ed->meta_event.txt);
				break;
			}
		}
		if (track_name[i] == NULL)
		{
			char foo[20];
			sprintf(foo,"TRACK_%d",i);
			track_name[i] = new UtilString(foo);
		}
	}
}

void ScoreCreationInfo::sort (MidiMsg** els, int len)
{
	if (len < 2)
		return;
	bool go_again = true;
	while (go_again)
	{
		go_again = false;
		for (int i = 0; i < len - 1; i++)
			if (els[i]->time > els[i+1]->time)
			{
				MidiMsg* temp = els[i];
				els[i] = els[i+1];
				els[i+1] = temp;
				go_again = true;
			}
	}
}

void ScoreCreationInfo::sort (FuncEl** els, int len)
{
	if (len < 2)
		return;
	bool go_again = true;
	while (go_again)
	{
		go_again = false;
		for (int i = 0; i < len - 1; i++)
			if (els[i]->time > els[i+1]->time)
			{
				FuncEl* temp = els[i];
				els[i] = els[i+1];
				els[i+1] = temp;
				go_again = true;
			}
	}
}

void ScoreCreationInfo::to_p(MidiMsgArrayPtrArrayPtr p, UINT numtracks)
{
	UINT i2 = 0;
	for (UINT i = 0; i < track_included.length; i++)
	{
		if (!track_included[i])
			continue;
		ioref node = track_node[i];
		MidiMsgArrayPtr p2 = fd->getTrack(i);
		if (p2->length < 1)
			continue;

		set_node_and_sanity_check(p2,node);

		(*p)[i2++] = p2;
	}
	if (num_msgs > 0) // handle manually inserted MidiMsgs
	{
		MidiMsgArrayPtr p2 = new MidiMsgArray(num_msgs);
		for (UINT j = 0; j < p2->length; j++)
			(*p2)[j] = *(msgs[j]);
		(*p)[i2++] = p2;
	}
	if (i2 != numtracks)
		r->ew.w("Sanity check failure in sci_to_p()");
}

int ScoreCreationInfo::numTracks()
{
	int numtracks = 0;
	for (UINT i = 0; i < track_included.length; i++)
		if (track_included[i])
			numtracks++;
	if (num_msgs > 0)
		numtracks++;
	return numtracks;
}

}
// Author: Patrick Pelletier
