/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "Score_Players.h"
#include "Score.h"
#include "err_and_warn.h"
#include "TrackMerger.h"
#include "MIDIsvc.h"
#include "Recorder.h"
#include "ScoreCreationInfo.h"
namespace rogus
{


void Score::convertFuncTime(ScoreTimeArrayPtr stap, ScoreTime lastTime, long lastTick, long thisTick, int* func_index, long currentTempo, long division)
{
	UINT i;
	for (i = *func_index; i < stap->length; i++)
	{
		if (thisTick != -1 && (*stap)[i] > thisTick)
		{
			*func_index = i;
			return;
		}
		long old = (*stap)[i];
		double multiplier = double(currentTempo)/double(division)/double(500);
		(*stap)[i] = (ScoreTime) (lastTime + ((*stap)[i]-lastTick)*multiplier + 0.5);
	}
	*func_index = i;
}

void Score::convertTime(MidiMsgArrayPtr p, long division, ScoreTimeArrayPtr stap)
{
	if (p->length < 1)
		return;
	// change tick time into real time
	int func_index = 0;
	ScoreTime lastTime = 0;
	long lastTick = 0;
	long currentTempo = 500000;

	for (UINT i = 0; i < p->length; i++)
	{
		MidiMsg* m = &((*p)[i]);
		long thisTick = m->time;
		double multiplier = double(currentTempo)/double(division)/double(500);
		
		m->time = (long) (lastTime + (thisTick-lastTick)*multiplier + 0.5);
		
		if (m->is_meta(meta::tempo))
		{
			convertFuncTime(stap, lastTime, lastTick, thisTick, &func_index, currentTempo, division);
			long newTempo = m->ed->meta_event.tem;
			lastTime += (ScoreTime) ((thisTick-lastTick)*multiplier + 0.5);
			lastTick = thisTick;
			currentTempo = newTempo;
		}
	}
	convertFuncTime(stap, lastTime, lastTick, -1, &func_index, currentTempo, division);
}

void Score::read(const Filedata* f)
{
	int2 numtracks = f->getNumTracks();
	int2 division = f->getDivision();
	if (numtracks < 1)
	{
		msgs = new MidiMsgArray(0);
		return;
	}
	ioref node = r->o.open_output();
	MidiMsgArrayPtrArrayPtr p = new MidiMsgArrayPtrArray(numtracks);
	for (int i = 0; i < numtracks; i++)
	{
		MidiMsgArrayPtr p2 = f->getTrack(i);
		if (p2->length < 1)
			continue;
		set_node_and_sanity_check(p2,node);
		(*p)[i] = p2;
	}
	msgs = TrackMerger<MidiMsg>::merge(p); // merge all the tracks together
	funcs = new ScoreFuncArray(0);
	args = new VoidStarArray(0);
	times = new ScoreTimeArray(0);
	convertTime(msgs, division, times);
}

Score::Score(const ScoreCreationInfo* sci_const)
{
	ScoreCreationInfo* sci = (ScoreCreationInfo*) sci_const;
      // cast away constness because we maintain logical constness
   r = sci->r;
	sci->do_sort();
	int numtracks = sci->numTracks();
	int2 division = sci->division;
	if (numtracks < 1)
	{
		msgs = new MidiMsgArray(0);
		return;
	}
	MidiMsgArrayPtrArrayPtr p = new MidiMsgArrayPtrArray(numtracks);

	sci->to_p(p,numtracks);
	
	msgs = TrackMerger<MidiMsg>::merge(p); // merge all the tracks together
	funcs = new ScoreFuncArray(sci->num_funcs);
	args = new VoidStarArray(sci->num_funcs);
	times = new ScoreTimeArray(sci->num_funcs);
	int j;
	for (j = 0; j < sci->num_funcs; j++)
	{
		(*funcs)[j] = sci->funcs[j]->func;
		(*args)[j] = sci->funcs[j]->v;
		(*times)[j] = sci->funcs[j]->time;
	}
	convertTime(msgs, division, times);
}


ScoreTime Score::getLength() const
{
	if (msgs->length < 1)
		return 0;
	return (*msgs)[msgs->length-1].time; // return time of last event
}

static void iterator(ioref, int, bool* e, void*)
{
	*e = 0;
}


Score::ChannelUseMapPtr Score::getChannels() const
{
	if (u != NULL)
		return u;
	Score::ChannelUseMapPtr result = new Score::ChannelUseMap(r->o.howManyOutputNodes());
	result->iterate(iterator, NULL);  // initialize to zero
	if (msgs->length)
		for (UINT i = 0; i < msgs->length; i++)
		{
			MidiMsg* m = &((*msgs)[i]);
			m->sanityCheck();
			if (m->msgt() != MidiMsg::system)
				(*result)(m->node, m->chan()) = 1;
		}
	((Score*)this)->u = result; // cache result: logical, not physical constness
	return result;
}

Score::Score(Rogus_McBogus* in_r, const Recorder* rec)
{
   r = in_r;
	msgs = new MidiMsgArray(rec->current_index);
	if (rec->current_index == 0)
		return;
	Recorder::storage_group* p = rec->first;
	int offset = 0;
	long mintime = (*p)[0].time;
	for (UINT i = 0; i < msgs->length; i++) // copy contents of Recorder
	{
		(*msgs)[i] = (*p)[i-offset];
		(*msgs)[i].time -= mintime; // adjust time
		if (i-offset > Recorder::storage_group::group_size)
		{
			offset += Recorder::storage_group::group_size;
			p = p->next;
		}
	}
	funcs = new ScoreFuncArray(0);
	args = new VoidStarArray(0);
	times = new ScoreTimeArray(0);
}
}
// Author: Patrick Pelletier
